/**
   @file
   @ingroup nr
*/
#include <bpm/bpm_messages.h>
#include <bpm/bpm_nr.h>


/**
   Calculates the Fourier transform on a set of n real valued datapoints
   replaces this data (array data[1..n] by the positive frequency half of 
   its complex Fourier transform. The real valued first and last components 
   of the complex tranform are returned as elements data[1] and data[2]
   respectively, n MUST be a power of 2. This routines calculates the
   inverse transform of a complex data array if it is the transform of
   real data, result in this case must be multiplied with 2/n
   
   BM. 15.08.2006: added the 2^n check on n
   Compute the FFT of a real function. NR 12.3 pg513
   
   @param data the array with the data, which gets replaced by fft
   @param n length of the data, must be power of 2
   @param isign sign of the transform
   
   @return BPM_FAILURE upon failure, BPM_SUCCESS upon success
*/
int nr_realft( double data[], unsigned long n, int isign ) {
  
    
  unsigned long i, i1,i2,i3,i4,np3;
  double c1=0.5,c2,h1r,h1i,h2r,h2i;
  double wr,wi,wpr,wpi,wtemp,theta;

  if ( ! nr_is_pow2( n ) ) {
    bpm_error( "Data length is not power of 2 in nr_realft(...)", 
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  theta = PI/(double)(n>>1);

  if (isign == 1){
      c2 = -0.5;
      nr_four1(data,n>>1,1);
  } else {
    c2=0.5;
    theta = -theta;
  }

  wtemp=sin(0.5*theta);
  wpr = -2.0*wtemp*wtemp;
  wpi=sin(theta);
  wr=1.0+wpr;
  wi=wpi;
  np3=n+3;
  for(i=2;i<=(n>>2);i++)
    {
      i4=1+(i3=np3-(i2=1+(i1=i+i-1)));
      h1r=c1*(data[i1]+data[i3]);
      h1i=c1*(data[i2]-data[i4]);
      h2r = -c2*(data[i2]+data[i4]);
      h2i=c2*(data[i1]-data[i3]);
      data[i1]=h1r+wr*h2r-wi*h2i;
      data[i2]=h1i+wr*h2i+wi*h2r;
      data[i3]=h1r-wr*h2r+wi*h2i;
      data[i4] = -h1i+wr*h2i+wi*h2r;
      wr=(wtemp=wr)*wpr-wi*wpi+wr;
      wi=wi*wpr+wtemp*wpi+wi;
    }
  if (isign == 1)
    {
      data[1] = (h1r=data[1])+data[2];
      data[2] = h1r-data[2];
    }
  else
    {
      data[1]=c1*((h1r=data[1])+data[2]);
      data[2]=c1*(h1r-data[2]);

      nr_four1(data,n>>1,-1);
    }

  return BPM_SUCCESS;
}
    
